class UnknownImageReference(ValueError):
    '''
    Represents an error for invalid/unknown image reference.
    The first argument of this exception should be the reference given by the user.
    '''

    def __str__(self):
        return f'Unknown image reference: {self.args[0]}'


class UnknownImageRegistry(ValueError):
    '''
    Represents an error for invalid/unknown image registry.
    The first argument of this exception should be the registry given by the user.
    '''

    def __str__(self):
        return f'Unknown image registry: {self.args[0]}'


class AliasResolutionFailed(ValueError):
    '''
    Represents an alias resolution failure.
    The first argument of this exception should be the alias given by the user.
    '''

    def __str__(self):
        return f'Failed to resolve alias: {self.args[0]}'



import ipaddress
import json
import logging
import re
from typing import Any, Mapping, Sequence, Tuple, Union

import aiohttp
import yarl

from .logging import BraceStyleAdapter
from .etcd import (
    AsyncEtcd,
    quote as etcd_quote,
    unquote as etcd_unquote,
)
from .exception import UnknownImageRegistry

__all__ = (
    'default_registry',
    'default_repository',
    'login',
    'get_known_registries',
    'is_known_registry',
    'get_registry_info',
)

log = BraceStyleAdapter(logging.Logger('ai.backend.common.docker'))

default_registry = 'index.docker.io'
default_repository = 'lablup'


async def login(
        sess: aiohttp.ClientSession,
        registry_url: yarl.URL,
        credentials: dict,
        scope: str) -> dict:
    '''
    Authorize to the docker registry using the given credentials and token scope, and returns a set
    of required aiohttp.ClientSession.request() keyword arguments for further API requests.

    Some registry servers only rely on HTTP Basic Authentication without token-based access controls
    (usually via nginx proxy). We do support them also. :)
    '''
    if credentials.get('username') and credentials.get('password'):
        basic_auth = aiohttp.BasicAuth(
            credentials['username'], credentials['password'],
        )
    else:
        basic_auth = None
    realm = registry_url / 'token'  # fallback
    service = 'registry'            # fallback
    async with sess.get(registry_url / 'v2/', auth=basic_auth) as resp:
        ping_status = resp.status
        www_auth_header = resp.headers.get('WWW-Authenticate')
        if www_auth_header:
            match = re.search(r'realm="([^"]+)"', www_auth_header)
            if match:
                realm = match.group(1)
            match = re.search(r'service="([^"]+)"', www_auth_header)
            if match:
                service = match.group(1)
    if ping_status == 200:
        log.debug('docker-registry: {0} -> basic-auth', registry_url)
        return {'auth': basic_auth, 'headers': {}}
    elif ping_status == 404:
        raise RuntimeError(f'Unsupported docker registry: {registry_url}! '
                           '(API v2 not implemented)')
    elif ping_status == 401:
        params = {
            'scope': scope,
            'offline_token': 'true',
            'client_id': 'docker',
            'service': service,
        }
        async with sess.get(realm, params=params, auth=basic_auth) as resp:
            log.debug('docker-registry: {0} -> {1}', registry_url, realm)
            if resp.status == 200:
                data = json.loads(await resp.read())
                token = data.get('token', None)
                return {'auth': None, 'headers': {
                    'Authorization': f'Bearer {token}'
                }}
    raise RuntimeError('authentication for docker registry '
                       'f{registry_url} failed')


async def get_known_registries(etcd: AsyncEtcd) -> Mapping[str, yarl.URL]:
    pairs = await etcd.get_prefix('config/docker/registry/')
    results = {}
    for key, value in pairs:
        parts = key.split('/')
        if len(parts) == 4:
            name = etcd_unquote(parts[-1])
            results[name] = value
    return results


def is_known_registry(val: str,
                      known_registries: Union[Mapping[str, Any], Sequence[str]]):
    if val == default_registry:
        return True
    if known_registries and val in known_registries:
        return True
    try:
        url = yarl.URL('//' + val)
        if url.host and ipaddress.ip_address(url.host):
            return True
    except ValueError:
        pass
    return False


async def get_registry_info(etcd: AsyncEtcd, name: str) -> Tuple[yarl.URL, dict]:
    reg_path = f'config/docker/registry/{etcd_quote(name)}'
    item = await etcd.get_prefix_dict(reg_path)
    if not item:
        raise UnknownImageRegistry(name)
    registry_addr = item['']
    creds = {}
    username = item.get('username')
    if username is not None:
        creds['username'] = username
    password = item.get('password')
    if password is not None:
        creds['password'] = password
    return yarl.URL(registry_addr), creds



import asyncio
import argparse
from collections import namedtuple
import ipaddress
import pathlib
import socket
from typing import Tuple
import threading

try:
    import aiodns
    _aiodns_available = True
    _aiodns_ctx = threading.local()
    _aiodns_ctx.resolver = None
except ImportError:
    _aiodns_available = False
    _aiodns_ctx = None


def port_no(s: str) -> int:
    try:
        port = int(s)
        assert port > 0
        assert port < 65536
    except (ValueError, AssertionError):
        msg = f'{s!r} is not a valid port number.'
        raise argparse.ArgumentTypeError(msg)
    return port


def port_range(s: str) -> Tuple[int, int]:
    try:
        port_range = tuple(map(int, s.split('-')))
    except (TypeError, ValueError):
        msg = f'{s!r} should be a hyphen-separated pair of integers.'
        raise argparse.ArgumentTypeError(msg)
    if len(port_range) != 2:
        msg = f'{s!r} should have exactly two integers.'
        raise argparse.ArgumentTypeError(msg)
    if not (0 < port_range[0] < 65536):
        msg = f'{port_range[0]} is not a valid port number.'
        raise argparse.ArgumentTypeError(msg)
    if not (0 < port_range[1] < 65536):
        msg = f'{port_range[1]} is not a valid port number.'
        raise argparse.ArgumentTypeError(msg)
    if not (port_range[0] < port_range[1]):
        msg = f'{port_range[0]} should be less than {port_range[1]}.'
        raise argparse.ArgumentTypeError(msg)
    return port_range


def positive_int(s: str) -> int:
    try:
        val = int(s)
        assert val > 0
    except (ValueError, AssertionError):
        msg = f'{s!r} is not a positive integer.'
        raise argparse.ArgumentTypeError(msg)
    return val


def non_negative_int(s: str) -> int:
    try:
        val = int(s)
        assert val >= 0
    except (ValueError, AssertionError):
        msg = f'{s!r} is not a non-negative integer.'
        raise argparse.ArgumentTypeError(msg)
    return val


class HostPortPair(namedtuple('_HostPortPair', 'host port')):

    def __format__(self, spec):
        return self.__str__()

    def __str__(self):
        if isinstance(self.host, ipaddress.IPv6Address):
            return f'[{self.host}]:{self.port}'
        return f'{self.host}:{self.port}'

    def as_sockaddr(self):
        # Translate this to a tuple of host/port pair without hostname resolving.
        return str(self.host), self.port

    def resolve(self):
        if isinstance(self.host, ipaddress._BaseAddress):
            # Already resolved one.
            return self
        # Resolve now and return a new HostPortPair.
        addrs = socket.getaddrinfo(self.host, 80)
        ip = ipaddress.ip_address(addrs[0][4][0])
        return HostPortPair(ip, self.port)

    async def resolve_async(self):
        if isinstance(self.host, ipaddress._BaseAddress):
            return self
        loop = asyncio.get_event_loop()
        if _aiodns_available:
            if _aiodns_ctx.resolver is None:
                _aiodns_ctx.resolver = aiodns.DNSResolver(loop=loop)
            else:
                assert _aiodns_ctx.resolver.loop is loop
            result = await _aiodns_ctx.resolver.gethostbyname(self.host, 0)
            ip = ipaddress.ip_address(result.addresses[0])
            return HostPortPair(ip, self.port)
        else:
            addrs = await loop.getaddrinfo(self.host, 80)
            ip = ipaddress.ip_address(addrs[0][4][0])
            return HostPortPair(ip, self.port)


def host_port_pair(s: str) -> Tuple[ipaddress._BaseAddress, int]:
    pieces = s.rsplit(':', maxsplit=1)
    if len(pieces) == 1:
        msg = f'{s!r} should contain both IP address and port number.'
        raise argparse.ArgumentTypeError(msg)
    elif len(pieces) == 2:
        # strip potential brackets in IPv6 hostname-port strings (RFC 3986).
        host = pieces[0].strip('[]')
        try:
            host = ipaddress.ip_address(host)
        except ValueError:
            # Let it be just a hostname.
            host = host
        try:
            port = int(pieces[1])
            assert port > 0
            assert port < 65536
        except (ValueError, AssertionError):
            msg = f'{pieces[1]!r} is not a valid port number.'
            raise argparse.ArgumentTypeError(msg)
    return HostPortPair(host, port)


def ipaddr(s: str) -> ipaddress._BaseAddress:
    try:
        ip = ipaddress.ip_address(s.strip('[]'))
    except ValueError:
        msg = f'{s!r} is not a valid IP address.'
        raise argparse.ArgumentTypeError(msg)
    return ip


def path(val: str) -> pathlib.Path:
    if val is None:
        return None
    p = pathlib.Path(val)
    if not p.exists():
        msg = f'{val!r} is not a valid file/dir path.'
        raise argparse.ArgumentTypeError(msg)
    return p



import json
import logging
import os
import socket
import sys
from pathlib import Path

import aiodns

from .utils import curl

__all__ = (
    'detect_cloud',
    'current_provider',
    'get_instance_id',
    'get_instance_ip',
    'get_instance_type',
    'get_instance_region',
)

log = logging.getLogger(__name__)


def is_containerized() -> bool:
    '''
    Check if I am running inside a Linux container.
    '''
    try:
        cginfo = Path('/proc/self/cgroup').read_text()
        if '/docker/' in cginfo or '/lxc/' in cginfo:
            return True
    except IOError:
        return False


def detect_cloud() -> str:
    '''
    Detect the cloud provider where I am running on.
    '''
    # NOTE: Contributions are welcome!
    # Please add other cloud providers such as Rackspace, IBM BlueMix, etc.
    if sys.platform.startswith('linux'):
        # Google Cloud Platform or Amazon AWS (hvm)
        try:
            # AWS Nitro-based instances
            mb = Path('/sys/devices/virtual/dmi/id/board_vendor').read_text().lower()
            if 'amazon' in mb:
                return 'amazon'
        except IOError:
            pass
        try:
            bios = Path('/sys/devices/virtual/dmi/id/bios_version').read_text().lower()
            if 'google' in bios:
                return 'google'
            if 'amazon' in bios:
                return 'amazon'
        except IOError:
            pass
        # Microsoft Azure
        # https://gallery.technet.microsoft.com/scriptcenter/Detect-Windows-Azure-aed06d51
        # TODO: this only works with Debian/Ubuntu instances.
        # TODO: this does not work inside containers.
        try:
            dhcp = Path('/var/lib/dhcp/dhclient.eth0.leases').read_text()
            if 'unknown-245' in dhcp:
                return 'azure'
            # alternative method is to read /var/lib/waagent/GoalState.1.xml
            # but it requires sudo privilege.
        except IOError:
            pass
    else:
        log.warning('Cloud detection is implemented for Linux only yet.')
    return None


# Detect upon module load.
current_provider = detect_cloud()
if current_provider is None:
    log.info('Detected environment: on-premise setup')
    log.info('The agent node ID is set using the hostname.')
else:
    log.info(f'Detected environment: {current_provider} cloud')
    log.info('The agent node ID will follow the instance ID.')

_defined = False
get_instance_id = None
get_instance_ip = None
get_instance_type = None
get_instance_region = None


def _define_functions():
    global _defined
    global get_instance_id
    global get_instance_ip
    global get_instance_type
    global get_instance_region
    if _defined:
        return

    if current_provider == 'amazon':
        # ref: http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
        _metadata_prefix = 'http://169.254.169.254/latest/meta-data/'
        _dynamic_prefix = 'http://169.254.169.254/latest/dynamic/'

        async def _get_instance_id():
            return await curl(_metadata_prefix + 'instance-id',
                              lambda: f'i-{socket.gethostname()}')

        async def _get_instance_ip():
            return await curl(_metadata_prefix + 'local-ipv4',
                              '127.0.0.1')

        async def _get_instance_type():
            return await curl(_metadata_prefix + 'instance-type',
                              'unknown')

        async def _get_instance_region():
            doc = await curl(_dynamic_prefix + 'instance-identity/document')
            if doc is None:
                return 'amazon/unknown'
            region = json.loads(doc)['region']
            return f'amazon/{region}'

    elif current_provider == 'azure':
        # ref: https://docs.microsoft.com/azure/virtual-machines/virtual-machines-instancemetadataservice-overview
        _metadata_prefix = 'http://169.254.169.254/metadata/instance'

        async def _get_instance_id():
            data = await curl(_metadata_prefix, None,
                              params={'version': '2017-03-01'},
                              headers={'Metadata': 'true'})
            if data is None:
                return f'i-{socket.gethostname()}'
            o = json.loads(data)
            return o['compute']['vmId']

        async def _get_instance_ip():
            data = await curl(_metadata_prefix, None,
                              params={'version': '2017-03-01'},
                              headers={'Metadata': 'true'})
            if data is None:
                return '127.0.0.1'
            o = json.loads(data)
            return o['network']['interface'][0]['ipv4']['ipaddress'][0]['ipaddress']

        async def _get_instance_type():
            data = await curl(_metadata_prefix, None,
                              params={'version': '2017-03-01'},
                              headers={'Metadata': 'true'})
            if data is None:
                return 'unknown'
            o = json.loads(data)
            return o['compute']['vmSize']

        async def _get_instance_region():
            data = await curl(_metadata_prefix, None,
                              params={'version': '2017-03-01'},
                              headers={'Metadata': 'true'})
            if data is None:
                return 'azure/unknown'
            o = json.loads(data)
            region = o['compute']['location']
            return f'azure/{region}'

    elif current_provider == 'google':
        # ref: https://cloud.google.com/compute/docs/storing-retrieving-metadata
        _metadata_prefix = 'http://metadata.google.internal/computeMetadata/v1/'

        async def _get_instance_id():
            return await curl(_metadata_prefix + 'instance/id',
                              lambda: f'i-{socket.gethostname()}',
                              headers={'Metadata-Flavor': 'Google'})

        async def _get_instance_ip():
            return await curl(_metadata_prefix + 'instance/network-interfaces/0/ip',
                              '127.0.0.1',
                              headers={'Metadata-Flavor': 'Google'})

        async def _get_instance_type():
            return await curl(_metadata_prefix + 'instance/machine-type',
                              'unknown',
                              headers={'Metadata-Flavor': 'Google'})

        async def _get_instance_region():
            zone = await curl(_metadata_prefix + 'instance/zone',
                              'unknown',
                              headers={'Metadata-Flavor': 'Google'})
            region = zone.rsplit('-', 1)[0]
            return f'google/{region}'

    else:
        _metadata_prefix = None

        async def _get_instance_id():
            return f'i-{socket.gethostname()}'

        async def _get_instance_ip():
            try:
                myself = socket.gethostname()
                resolver = aiodns.DNSResolver()
                result = await resolver.gethostbyname(myself, socket.AF_INET)
                return result.addresses[0]
            except aiodns.error.DNSError:
                return '127.0.0.1'

        async def _get_instance_type():
            return 'default'

        async def _get_instance_region():
            return os.environ.get('BACKEND_REGION', 'local')

    get_instance_id = _get_instance_id
    get_instance_ip = _get_instance_ip
    get_instance_type = _get_instance_type
    get_instance_region = _get_instance_region
    _defined = True


_define_functions()



from abc import ABC, abstractmethod


class AbstractStatsMonitor(ABC):

    @abstractmethod
    def report_stats(self, report_type, metric, *args):
        raise NotImplementedError

    @abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


class AbstractErrorMonitor(ABC):

    @abstractmethod
    def capture_exception(self, *args):
        raise NotImplementedError

    @abstractmethod
    def capture_message(self, msg):
        raise NotImplementedError

    @abstractmethod
    def set_context(self, context):
        raise NotImplementedError

    @abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


'''
A collection of dummy monitoring daemon classes with no-op stubs.
Useful to avoid excessive repetition of "if app['stats_monitor']:
app['stats_monitor'].statsd.increment(...)" snippets all scattered around the
codebase.
'''


class DummyStatsMonitor(AbstractStatsMonitor):
    def report_stats(self, *args, **kwargs): pass        # noqa
    def __enter__(self): return self                     # noqa
    def __exit__(self, exc_type, exc_val, exc_tb): pass  # noqa


class DummyErrorMonitor(AbstractErrorMonitor):
    def capture_exception(self, *args, **kwargs): pass   # noqa
    def capture_message(self, *args, **kwargs): pass     # noqa
    def set_context(self, context): pass                 # noqa
    def __enter__(self): return self                     # noqa
    def __exit__(self, exc_type, exc_val, exc_tb): pass  # noqa



from collections import OrderedDict
from datetime import datetime
import itertools
import logging, logging.config, logging.handlers
import multiprocessing as mp
import os
from pathlib import Path
import signal
import urllib.request, urllib.error

from setproctitle import setproctitle
from pythonjsonlogger.jsonlogger import JsonFormatter
import zmq

__all__ = ('Logger', 'BraceStyleAdapter')

_logging_ctx = zmq.Context()


class LogstashHandler(logging.Handler):

    def __init__(self, endpoint):
        super().__init__()
        self.endpoint = endpoint
        self._cached_priv_ip = None
        self._cached_inst_id = None
        self._ctx = _logging_ctx
        self._sock = self._ctx.socket(zmq.PUB)
        self._sock.setsockopt(zmq.LINGER, 50)
        self._sock.setsockopt(zmq.SNDHWM, 20)
        self._sock.connect(self.endpoint)

    def emit(self, record):
        now = datetime.now().isoformat()
        tags = set()
        extra_data = dict()

        if record.exc_info:
            tags.add('has_exception')
            if self.formatter:
                extra_data['exception'] = self.formatter.formatException(record.exc_info)
            else:
                extra_data['exception'] = logging._defaultFormatter.formatException(record.exc_info)

        # This log format follows logstash's event format.
        log = OrderedDict([
            ('@timestamp', now),
            ('@version', 1),
            ('host', self._get_my_private_ip()),
            ('instance', self._get_my_instance_id()),
            ('logger', record.name),
            ('loc', '{0}:{1}:{2}'.format(record.pathname, record.funcName, record.lineno)),
            ('message', record.getMessage()),
            ('level', record.levelname),
            ('tags', list(tags)),
        ])
        log.update(extra_data)
        self._sock.send_json(log)

    def _get_my_private_ip(self):
        # See the full list of EC2 instance metadata at
        # http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/ec2-instance-metadata.html
        if self._cached_priv_ip is None:
            try:
                resp = urllib.request.urlopen('http://169.254.169.254/latest/meta-data/local-ipv4', timeout=0.2)
                self._cached_priv_ip = resp.read().decode('ascii')
            except urllib.error.URLError:
                self._cached_priv_ip = 'unavailable'
        return self._cached_priv_ip

    def _get_my_instance_id(self):
        if self._cached_inst_id is None:
            try:
                resp = urllib.request.urlopen('http://169.254.169.254/latest/meta-data/instance-id', timeout=0.2)
                self._cached_inst_id = resp.read().decode('ascii')
            except urllib.error.URLError:
                self._cached_inst_id = 'i-00000000'
        return self._cached_inst_id


class CustomJsonFormatter(JsonFormatter):

    def add_fields(self, log_record, record, message_dict):
        super().add_fields(log_record, record, message_dict)
        if not log_record.get('timestamp'):
            # this doesn't use record.created, so it is slightly off
            now = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            log_record['timestamp'] = now
        if log_record.get('level', record.levelname):
            log_record['level'] = log_record['level'].upper()
        else:
            log_record['level'] = record.levelname


def log_worker(config, parent_pid, log_queue):
    setproctitle(f'backend.ai: logger pid({parent_pid})')
    if config.log_file is not None:
        fmt = '(timestamp) (level) (name) (processName) (message)'
        file_handler = logging.handlers.RotatingFileHandler(
            filename=config.log_file,
            backupCount=config.log_file_count,
            maxBytes=1048576 * config.log_file_size,
            encoding='utf-8',
        )
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(CustomJsonFormatter(fmt))
    while True:
        rec = log_queue.get()
        if rec is None:
            break
        file_handler.emit(rec)


class BraceMessage:

    __slots__ = ('fmt', 'args')

    def __init__(self, fmt, args):
        self.fmt = fmt
        self.args = args

    def __str__(self):
        return self.fmt.format(*self.args)


class BraceStyleAdapter(logging.LoggerAdapter):

    def __init__(self, logger, extra=None):
        super().__init__(logger, extra)

    def log(self, level, msg, *args, **kwargs):
        if self.isEnabledFor(level):
            msg, kwargs = self.process(msg, kwargs)
            self.logger._log(level, BraceMessage(msg, args), (), **kwargs)


class Logger():

    def __init__(self, config):
        self.log_config = {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'colored': {
                    '()': 'coloredlogs.ColoredFormatter',
                    'format': '%(asctime)s %(levelname)s %(name)s '
                              '[%(process)d] %(message)s',
                    'field_styles': {'levelname': {'color': 248, 'bold': True},
                                     'name': {'color': 246, 'bold': False},
                                     'process': {'color': 'cyan'},
                                     'asctime': {'color': 240}},
                    'level_styles': {'debug': {'color': 'green'},
                                     'verbose': {'color': 'green', 'bright': True},
                                     'info': {'color': 'cyan', 'bright': True},
                                     'notice': {'color': 'cyan', 'bold': True},
                                     'warning': {'color': 'yellow'},
                                     'error': {'color': 'red', 'bright': True},
                                     'success': {'color': 77},
                                     'critical': {'background': 'red', 'color': 255, 'bold': True}},
                },
            },
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG',
                    'formatter': 'colored',
                    'stream': 'ext://sys.stderr',
                },
                'null': {
                    'class': 'logging.NullHandler',
                },
            },
            'loggers': {
                '': {
                    'handlers': ['console'],
                    'level': 'INFO',
                },
            },
        }
        self.cli_config = config

    @staticmethod
    def update_log_args(parser):
        parser.add('--debug', env_var='BACKEND_DEBUG',
                   action='store_true', default=False,
                   help='Set the debug mode and verbose logging. (default: false)')
        parser.add('-v', '--verbose', env_var='BACKEND_VERBOSE',
                   action='store_true', default=False,
                   help='Set even more verbose logging which includes all SQL '
                        'statements issued. (default: false)')
        parser.add('--log-file', env_var='BACKEND_LOG_FILE',
                   type=Path, default=None,
                   help='If set to a file path, line-by-line JSON logs will be '
                        'recorded there.  It also automatically rotates the logs '
                        'using dotted number suffixes. (default: None)')
        parser.add('--log-file-count', env_var='BACKEND_LOG_FILE_COUNT',
                   type=int, default=10,
                   help='The maximum number of rotated log files (default: 10)')
        parser.add('--log-file-size', env_var='BACKEND_LOG_FILE_SIZE',
                   type=float, default=10.0,
                   help='The maximum size of each log file in MiB '
                        '(default: 10 MiB)')

    def add_pkg(self, pkgpath):
        self.log_config['loggers'][pkgpath] = {
            'handlers': ['console'],
            'propagate': False,
            'level': 'DEBUG' if self.cli_config.debug else 'INFO',
        }

    def __enter__(self):
        self.log_queue = mp.Queue()
        if self.cli_config.log_file is not None:
            self.log_config['handlers']['fileq'] = {
                'class': 'logging.handlers.QueueHandler',
                'level': 'DEBUG',
                'queue': self.log_queue,
            }
            for l in self.log_config['loggers'].values():
                l['handlers'].append('fileq')
        logging.config.dictConfig(self.log_config)
        # block signals that may interrupt/corrupt logging
        stop_signals = {signal.SIGINT, signal.SIGTERM}
        signal.pthread_sigmask(signal.SIG_BLOCK, stop_signals)
        proc = mp.Process(target=log_worker, name='Logger',
                          args=(self.cli_config, os.getpid(), self.log_queue))
        proc.start()
        signal.pthread_sigmask(signal.SIG_UNBLOCK, stop_signals)
        # reset process counter
        mp.process._process_counter = itertools.count(0)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.log_queue.put(None)
        self.log_queue.close()
        self.log_queue.join_thread()



'''
An asynchronous client wrapper for etcd v3 API.

It uses the etcd3 library using a thread pool executor.
We plan to migrate to aioetcd3 library but it requires more work to get maturity.
Fortunately, etcd3's watchers are not blocking because they are implemented
using callbacks in separate threads.
'''

import asyncio
import collections
from concurrent.futures import ThreadPoolExecutor
import functools
import logging
from typing import Any, Mapping
from urllib.parse import quote as _quote, unquote

import etcd3

__all__ = (
    'quote', 'unquote',
    'AsyncEtcd',
)

Event = collections.namedtuple('Event', 'key event value')
log = logging.getLogger(__name__)

quote = functools.partial(_quote, safe='')


def make_dict_from_pairs(key_prefix, pairs, path_sep='/'):
    result = {}
    len_prefix = len(key_prefix)
    if isinstance(pairs, dict):
        iterator = pairs.items()
    else:
        iterator = pairs
    for k, v in iterator:
        if not k.startswith(key_prefix):
            continue
        subkey = k[len_prefix:]
        if subkey.startswith(path_sep):
            subkey = subkey[1:]
        path_components = subkey.split('/')
        parent = result
        for p in path_components[:-1]:
            if p not in parent:
                parent[p] = {}
            if p in parent and not isinstance(parent[p], dict):
                root = parent[p]
                parent[p] = {'': root}
            parent = parent[p]
        parent[path_components[-1]] = v
    return result


class AsyncEtcd:

    def __init__(self, addr, namespace, *, encoding='utf8', loop=None):
        self.loop = loop if loop else asyncio.get_event_loop()
        self.executor = ThreadPoolExecutor(max_workers=5, thread_name_prefix='etcd')
        self.etcd_sync = etcd3.client(host=str(addr.host), port=addr.port)
        self.ns = namespace
        log.info(f'using etcd cluster from {addr} with namespace "{namespace}"')
        self.encoding = encoding

    def _mangle_key(self, k):
        if k.startswith('/'):
            k = k[1:]
        return f'/sorna/{self.ns}/{k}'.encode(self.encoding)

    def _demangle_key(self, k):
        if isinstance(k, bytes):
            k = k.decode(self.encoding)
        prefix = f'/sorna/{self.ns}/'
        if k.startswith(prefix):
            k = k[len(prefix):]
        return k

    async def put(self, key, val):
        key = self._mangle_key(key)
        return await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.put, key, str(val).encode(self.encoding))

    async def put_multi(self, keys, values):
        return await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.transaction,
            [],
            [self.etcd_sync.transactions.put(
                self._mangle_key(k), str(v).encode(self.encoding))
             for k, v in zip(keys, values)],
            [])

    async def put_dict(self, dict_obj: Mapping[str, Any]):
        return await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.transaction,
            [],
            [self.etcd_sync.transactions.put(
                self._mangle_key(k), str(v).encode(self.encoding))
             for k, v in dict_obj.items()],
            [])

    async def get(self, key):
        key = self._mangle_key(key)
        val, _ = await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.get, key)
        return val.decode(self.encoding) if val is not None else None

    async def get_prefix(self, key_prefix):
        key_prefix = self._mangle_key(key_prefix)
        results = await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.get_prefix, key_prefix)
        return ((self._demangle_key(t[1].key),
                 t[0].decode(self.encoding))
                for t in results)

    async def get_prefix_dict(self, key_prefix, path_sep='/'):
        pairs = await self.get_prefix(key_prefix)
        return make_dict_from_pairs(key_prefix, pairs, path_sep)

    async def replace(self, key, initial_val, new_val):
        key = self._mangle_key(key)
        success = await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.replace, key, initial_val, new_val)
        return success

    async def delete(self, key):
        key = self._mangle_key(key)
        return await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.delete, key)

    async def delete_multi(self, keys):
        return await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.transaction,
            [],
            [self.etcd_sync.transactions.delete(self._mangle_key(k))
             for k in keys],
            [])

    async def delete_prefix(self, key_prefix):
        key_prefix = self._mangle_key(key_prefix)
        return await self.loop.run_in_executor(
            self.executor,
            self.etcd_sync.delete_prefix, key_prefix)

    def _watch_cb(self, queue, ev):
        if isinstance(ev, etcd3.events.PutEvent):
            ev_type = 'put'
        elif isinstance(ev, etcd3.events.DeleteEvent):
            ev_type = 'delete'
        else:
            assert False, 'Not recognized etcd event type.'
        # etcd3 library uses a separate thread for its watchers.
        event = Event(
            self._demangle_key(ev.key),
            ev_type,
            ev.value.decode(self.encoding),
        )
        self.loop.call_soon_threadsafe(queue.put_nowait, event)

    async def _watch_impl(self, raw_key, **kwargs):
        queue = asyncio.Queue(loop=self.loop)
        cb = functools.partial(self._watch_cb, queue)
        watch_id = self.etcd_sync.add_watch_callback(raw_key, cb, **kwargs)
        try:
            while True:
                ev = await queue.get()
                yield ev
                queue.task_done()
        except asyncio.CancelledError:
            pass
        finally:
            self.etcd_sync.cancel_watch(watch_id)
            del queue

    async def watch(self, key, *, once=False, ready_event=None):
        key = self._mangle_key(key)
        if ready_event:
            ready_event.set()
        # yield from in async-generator is not supported.
        async for ev in self._watch_impl(key):
            yield ev
            if once:
                break

    async def watch_prefix(self, key_prefix, *, once=False, ready_event=True):
        key_prefix = self._mangle_key(key_prefix)
        range_end = etcd3.utils.increment_last_byte(etcd3.utils.to_bytes(key_prefix))
        if ready_event:
            ready_event.set()
        async for ev in self._watch_impl(key_prefix, range_end=range_end):
            yield ev
            if once:
                break



import asyncio
import inspect
import logging
import pkg_resources
import typing

from .logging import BraceStyleAdapter
from .monitor import AbstractErrorMonitor, AbstractStatsMonitor

log = BraceStyleAdapter(logging.getLogger('ai.backend.common.plugin'))


plugin_base_classes = {
    'stats_monitor': AbstractStatsMonitor,
    'error_monitor': AbstractErrorMonitor,
}


class PluginRegistry:

    def __init__(self, plugin_name):
        self._plugins = []
        self._base_class = plugin_base_classes[plugin_name]
        self._methods = {}
        methods = inspect.getmembers(self._base_class,
                                     predicate=inspect.ismethod)
        for name, method in methods:
            self._methods[name] = method

    def register(self, plugin):
        assert isinstance(plugin, self._base_class), \
            (f'Wrong type of plugin '
             f'(plugin: {type(plugin)} / registry: {self._base_class})')
        self._plugins.append(plugin)

    def __getattr__(self, name):
        try:
            method = self._methods[name]
        except KeyError:
            raise AttributeError(f"'{self._base_class.__name__}' object "
                                 f"has no attribute '{name}'")

        if inspect.iscoroutinefunction(method):
            async def _callback_dispatcher(*args, **kwargs):
                return await asyncio.gather(*[getattr(plugin, name)(*args, kwargs)
                                              for plugin in self._plugins],
                                            return_exceptions=True)
        else:
            def _callback_dispatcher(*args, **kwargs):
                results = []
                for plugin in self._plugins:
                    result = getattr(plugin, name)(*args, **kwargs)
                    results.append(result)
                return results

        return _callback_dispatcher


def discover_entrypoints(plugins, disable_plugins=None):
    if disable_plugins is None:
        disable_plugins = []
    for plugin_name in plugins:
        plugin_group = f'backendai_{plugin_name}_v10'
        for entrypoint in pkg_resources.iter_entry_points(plugin_group):
            if entrypoint.name in disable_plugins:
                continue
            yield plugin_group, plugin_name, entrypoint


def install_plugins(plugins, app, install_type, config):
    """
    Automatically install plugins to the app.

    :param plugins: List of plugin names to discover and install plugins
    :param app: Any type of app to install plugins
    :param install_type: The way to install plugins to app
    :param config: Config object to initialize plugins
    :return:

    You should note that app can be any type of object. For instance,
    when used in manager, app param is the instance of aiohttp.web.Application,
    but it is the instance of subclass of aiozmq.rpc.AttrHandler in agents.

    Therefore, you should specify :install_type: to install plugins into different
    types of apps correctly. Currently we support two types of :install_type:,
    which are 'attr' and 'dict'. For 'attr', plugins will be installed to app
    as its attributes. For 'dict', plugins will be installed as following:
    app[plugin_name] = plugin.
    """
    try:
        disable_plugins = config.disable_plugins
        if not disable_plugins:
            disable_plugins = []
    except AttributeError:
        disable_plugins = []
    for plugin_name in plugins:
        plugin_group = f'backendai_{plugin_name}_v10'
        registry = PluginRegistry(plugin_name)
        for entrypoint in pkg_resources.iter_entry_points(plugin_group):
            if entrypoint.name in disable_plugins:
                continue
            log.info('Installing plugin: {}.{}', plugin_group, entrypoint.name)
            plugin_module = entrypoint.load()
            plugin = getattr(plugin_module, 'get_plugin')(config)
            registry.register(plugin)
            if install_type == 'attr':
                setattr(app, plugin_name, registry)
            elif install_type == 'dict':
                assert isinstance(app, typing.MutableMapping), \
                    (f"app must be an instance of MutableMapping "
                     f"for 'dict' install_type.")
                app[plugin_name] = registry
            else:
                raise ValueError(f'Invalid install type: {install_type}')


def add_plugin_args(parser, plugins):
    for plugin_name in plugins:
        plugin_group = f'backendai_{plugin_name}_v10'
        for entrypoint in pkg_resources.iter_entry_points(plugin_group):
            plugin_module = entrypoint.load()
            _add_plugin_args = getattr(plugin_module, 'add_plugin_args', None)
            if _add_plugin_args:
                _add_plugin_args(parser)



from collections import UserDict
from decimal import Decimal
import enum
import numbers
from packaging import version
import re
from typing import (
    Any, Hashable, Mapping,
    Iterable, Sequence, Set,
    NewType, Tuple, Union
)

import attr
import itertools

from . import etcd
from .docker import (
    default_registry, default_repository,
    is_known_registry, get_known_registries,
)
from .exception import AliasResolutionFailed

__all__ = (
    'BinarySize',
    'ImageRef',
    'PlatformTagSet',
    'DeviceId',
    'SlotType',
    'IntrinsicSlotTypes',
    'Allocation',
    'ResourceAllocations',
    'MountPermission',

    # TODO: to be updated
    'ShareRequest',
    'SessionRequest',
)


class IntrinsicSlotTypes(str, enum.Enum):
    CPU = 'cpu'
    MEMORY = 'mem'


DeviceId = NewType('DeviceId', Hashable)
SlotType = NewType('DeviceType', Union[str, IntrinsicSlotTypes])


class MountPermission(str, enum.Enum):
    READ_ONLY = 'ro'
    READ_WRITE = 'rw'
    WRITE_DELETE = 'wd'


class BinarySize(int):
    '''
    A wrapper around Python integers to represent binary sizes for storage and
    memory in various places.

    Its string representation and parser, ``from_str()`` classmethod, does not use
    any locale-specific digit delimeters -- it supports only standard Python
    digit delimeters.
    '''

    suffix_map = {
        'y': 2 ** 80, 'Y': 2 ** 80,  # yotta
        'z': 2 ** 70, 'Z': 2 ** 70,  # zetta
        'e': 2 ** 60, 'E': 2 ** 60,  # exa
        'p': 2 ** 50, 'P': 2 ** 50,  # peta
        't': 2 ** 40, 'T': 2 ** 40,  # tera
        'g': 2 ** 30, 'G': 2 ** 30,  # giga
        'm': 2 ** 20, 'M': 2 ** 20,  # mega
        'k': 2 ** 10, 'K': 2 ** 10,  # kilo
        ' ': 1,
    }
    suffices = (' ', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    endings = ('ibytes', 'ibyte', 'ib', 'bytes', 'byte', 'b')

    @classmethod
    def from_str(cls, expr):
        if isinstance(expr, numbers.Integral):
            return cls(expr)
        assert isinstance(expr, str)
        orig_expr = expr
        expr = expr.strip().replace('_', '')
        try:
            return int(expr)
        except ValueError:
            expr = expr.lower()
            for ending in cls.endings:
                if expr.endswith(ending):
                    length = len(ending) + 1
                    suffix = expr[-length]
                    expr = Decimal(expr[:-length])
                    break
            else:
                # when there is no unit ending (e.g., "2K")
                if not str.isnumeric(expr[-1]):
                    suffix = expr[-1]
                    expr = Decimal(expr[:-1])
                else:
                    suffix = ' '
            try:
                multiplier = cls.suffix_map[suffix]
                return cls(expr * multiplier)
            except KeyError:
                raise ValueError('Unconvertible value', orig_expr)

    def _preformat(self):
        scale = self
        suffix_idx = 0
        while scale >= 1024:
            scale //= 1024
            suffix_idx += 1
        return suffix_idx

    @staticmethod
    def _quantize(val, multiplier):
        d = Decimal(val) / Decimal(multiplier)
        if d == d.to_integral():
            value = d.quantize(Decimal(1))
        else:
            value = d.quantize(Decimal('.00')).normalize()
        return value

    def __str__(self):
        suffix_idx = self._preformat()
        if suffix_idx == 0:
            if self == 1:
                return f'{int(self)} byte'
            else:
                return f'{int(self)} bytes'
        else:
            suffix = type(self).suffices[suffix_idx]
            multiplier = type(self).suffix_map[suffix]
            value = self._quantize(self, multiplier)
            return f'{value} {suffix.upper()}iB'

    def __format__(self, format_spec):
        if len(format_spec) != 1:
            raise ValueError('format-string for BinarySize can be only one character.')
        if format_spec == 's':
            # automatically scaled
            suffix_idx = self._preformat()
            if suffix_idx == 0:
                return f'{int(self)}'
            suffix = type(self).suffices[suffix_idx]
            multiplier = type(self).suffix_map[suffix]
            value = self._quantize(self, multiplier)
            return f'{value}{suffix.lower()}'
        else:
            # use the given scale
            suffix = format_spec.lower()
            multiplier = type(self).suffix_map.get(suffix)
            if multiplier is None:
                raise ValueError('Unsupported scale unit.', suffix)
            value = self._quantize(self, multiplier)
            return f'{value}{suffix.lower()}'.strip()
        return super().__format__(format_spec)


Allocation = NewType('Allocation', Union[Decimal, BinarySize])

ResourceAllocations = NewType('ResourceAllocations',
    Mapping[SlotType, Mapping[DeviceId, Allocation]])
'''
Represents the mappings of resource slot types and
their device-allocation pairs.

Example:

 + "cpu"
   - "0": 1.0
   - "1": 1.0
 + "mem"
   - "node0": "4g"
   - "node1": "4g"
 + "cuda.smp"
   - "0": 20
   - "1": 10
 + "cuda.mem"
   - "0": "8g"
   - "1": "2g"
'''


class PlatformTagSet(Mapping):

    __slots__ = ('_data', )
    _rx_ver = re.compile(r'^(?P<tag>[a-zA-Z]+)(?P<version>\d+(?:\.\d+)*[a-z0-9]*)?$')

    def __init__(self, tags: Iterable[str]):
        self._data = dict()
        rx = type(self)._rx_ver
        for t in tags:
            match = rx.search(t)
            if match is None:
                raise ValueError('invalid tag-version string', t)
            key = match.group('tag')
            value = match.group('version')
            if key in self._data:
                raise ValueError('duplicate platform tag with different versions', t)
            if value is None:
                value = ''
            self._data[key] = value

    def has(self, key: str, version: str = None):
        if version is None:
            return key in self._data
        _v = self._data.get(key, None)
        return _v == version

    def __getitem__(self, key: str):
        return self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __eq__(self, other):
        if isinstance(other, (set, frozenset)):
            return set(self._data.keys()) == other
        return self._data == other


class ImageRef:

    __slots__ = ('_registry', '_name', '_tag', '_tag_set', '_sha')

    _rx_slug = re.compile(r'^[A-Za-z0-9](?:[A-Za-z0-9-._]*[A-Za-z0-9])?$')

    @classmethod
    async def resolve_alias(cls, alias_key: str, etcd: etcd.AsyncEtcd):
        '''
        Resolve the tag using etcd so that the current instance indicates
        a concrete, latest image.

        Note that alias resolving does not take the registry component into
        account.
        '''
        alias_target = None
        repeats = 0
        while repeats < 8:
            prev_alias_key = alias_key
            alias_key = await etcd.get(f'images/_aliases/{alias_key}')
            if alias_key is None:
                alias_target = prev_alias_key
                break
            repeats += 1
        else:
            raise AliasResolutionFailed('Could not resolve the given image name!')
        known_registries = await get_known_registries(etcd)
        return cls(alias_target, known_registries)

    def __init__(self, value: str,
                 known_registries: Union[Mapping[str, Any], Sequence[str]] = None):
        rx_slug = type(self)._rx_slug
        if '://' in value or value.startswith('//'):
            raise ValueError('ImageRef should not contain the protocol scheme.')
        parts = value.split('/', maxsplit=1)
        if len(parts) == 1:
            self._registry = default_registry
            self._name, self._tag = ImageRef._parse_image_tag(value, True)
            if not rx_slug.search(self._tag):
                raise ValueError('Invalid image tag')
        else:
            if is_known_registry(parts[0], known_registries):
                self._registry = parts[0]
                using_default = (parts[0].endswith('.docker.io') or parts[0] == 'docker.io')
                self._name, self._tag = ImageRef._parse_image_tag(parts[1], using_default)
            else:
                self._registry = default_registry
                self._name, self._tag = ImageRef._parse_image_tag(value, True)
            if not rx_slug.search(self._tag):
                raise ValueError('Invalid image tag')
        self._update_tag_set()

    @staticmethod
    def _parse_image_tag(s: str, using_default_registry: bool = False) -> Tuple[str, str]:
        image_tag = s.rsplit(':', maxsplit=1)
        if len(image_tag) == 1:
            image = image_tag[0]
            tag = 'latest'
        else:
            image = image_tag[0]
            tag = image_tag[1]
        if not image:
            raise ValueError('Empty image repository/name')
        if ('/' not in image) and using_default_registry:
            image = default_repository + '/' + image
        return image, tag

    def _update_tag_set(self):
        if self._tag is None:
            self._tag_set = (None, PlatformTagSet([]))
            return
        tags = self._tag.split('-')
        self._tag_set = (tags[0], PlatformTagSet(tags[1:]))

    def generate_aliases(self) -> Mapping[str, 'ImageRef']:
        basename = self.name.split('/')[-1]
        possible_names = basename.rsplit('-')
        if len(possible_names) > 1:
            possible_names = [basename, possible_names[1]]

        possible_ptags = []
        tag_set = self.tag_set
        if not tag_set[0]:
            pass
        else:
            possible_ptags.append([tag_set[0]])
            for tag_key in tag_set[1]:
                tag_ver = tag_set[1][tag_key]
                tag_list = ['', tag_key, tag_key + tag_ver]
                if '.' in tag_ver:
                    tag_list.append(tag_key + tag_ver.rsplit('.')[0])
                elif tag_key == 'py' and len(tag_ver) > 1:
                    tag_list.append(tag_key + tag_ver[0])
                if 'cuda' in tag_key:
                    tag_list.append('gpu')
                possible_ptags.append(tag_list)

        ret = {}
        for name in possible_names:
            ret[name] = self
        for name, ptags in itertools.product(
                possible_names,
                itertools.product(*possible_ptags)):
            ret[f"{name}:{'-'.join(t for t in ptags if t)}"] = self
        return ret

    @staticmethod
    def merge_aliases(genned_aliases_1, genned_aliases_2) -> Mapping[str, 'ImageRef']:
        ret = {}
        aliases_set_1, aliases_set_2 = set(genned_aliases_1.keys()), set(genned_aliases_2.keys())
        aliases_dup = aliases_set_1 & aliases_set_2

        for alias in aliases_dup:
            ret[alias] = max(genned_aliases_1[alias], genned_aliases_2[alias])

        for alias in aliases_set_1 - aliases_dup:
            ret[alias] = genned_aliases_1[alias]
        for alias in aliases_set_2 - aliases_dup:
            ret[alias] = genned_aliases_2[alias]

        return ret

    @property
    def canonical(self) -> str:
        # e.g., registry.docker.io/lablup/kernel-python:3.6-ubuntu
        return f'{self.registry}/{self.name}:{self.tag}'

    @property
    def tag_path(self) -> str:
        '''
        Return the string key that can be used to fetch image metadata from etcd.
        '''
        return f'images/{etcd.quote(self.registry)}/' \
               f'{etcd.quote(self.name)}/{self.tag}'

    @property
    def registry(self) -> str:
        # e.g., lablup
        return self._registry

    @property
    def name(self) -> str:
        # e.g., python
        return self._name

    @property
    def tag(self) -> str:
        # e.g., 3.6-ubuntu
        return self._tag

    @property
    def tag_set(self) -> Tuple[str, PlatformTagSet]:
        # e.g., '3.6', {'ubuntu', 'cuda', ...}
        return self._tag_set

    @property
    def short(self) -> str:
        '''
        Returns the image reference string without the registry part.
        '''
        # e.g., python:3.6-ubuntu
        return f'{self.name}:{self.tag}' if self.tag is not None else self.name

    def __str__(self) -> str:
        return self.canonical

    def __repr__(self) -> str:
        return f'<ImageRef: "{self.canonical}">'

    def __hash__(self) -> int:
        return hash((self._name, self._tag, self._registry))

    def __lt__(self, other) -> bool:
        if self == other:   # call __eq__ first for resolved check
            return False
        if self.name != other.name:
            raise ValueError('only the image-refs with same names can be compared.')
        if self.tag_set[0] != other.tag_set[0]:
            return version.parse(self.tag_set[0]) < version.parse(other.tag_set[0])
        ptagset_self, ptagset_other = self.tag_set[1], other.tag_set[1]
        for key_self in ptagset_self:
            if ptagset_other.has(key_self):
                version_self, version_other = ptagset_self.get(key_self), ptagset_other.get(key_self)
                if version_self and version_other:
                    parsed_version_self, parsed_version_other = version.parse(version_self), version.parse(version_other)
                    if parsed_version_self != parsed_version_other:
                        return parsed_version_self < parsed_version_other
        return len(ptagset_self) > len(ptagset_other)


class ResourceSlot(UserDict):

    __slots__ = ('data', 'numeric')

    def __init__(self, *args, numeric=False):
        super().__init__(*args)
        self.numeric = numeric

    def __add__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be added together.')
        return type(self)({
            k: self.get(k, 0) + other.get(k, 0)
            for k in (self.keys() | other.keys())
        }, numeric=True)

    def __sub__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be operands of subtraction.')
        if other.keys() - self.keys():
            raise ValueError('Cannot subtract resource slot with more keys!')
        return type(self)({
            k: self.data[k] - other.get(k, 0)
            for k in self.keys()
        }, numeric=True)

    def __eq__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() != other.keys():
            raise TypeError('Only slots with same keys can be compared.')
        self_values = [self.data[k] for k in sorted(self.data.keys())]
        other_values = [other.data[k] for k in sorted(other.data.keys())]
        return self_values == other_values

    def __ne__(self, other):
        return not self.__eq__(other)

    def eq_contains(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() < other.keys():
            raise ValueError('Slots with less keys cannot contain other.')
        common_keys = sorted(other.keys() & self.keys())
        self_values = [self.data[k] for k in common_keys]
        other_values = [other.data[k] for k in common_keys]
        return self_values == other_values

    def eq_contained(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() > other.keys():
            raise ValueError('Slots with more keys cannot be contained in other.')
        common_keys = sorted(other.keys() & self.keys())
        self_values = [self.data[k] for k in common_keys]
        other_values = [other.data[k] for k in common_keys]
        return self_values == other_values

    def __le__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() > other.keys():
            raise ValueError('Slots with more keys cannot be smaller than other.')
        self_values = [self.data[k] for k in self.keys()]
        other_values = [other.data[k] for k in self.keys()]
        return not any(s > o for s, o in zip(self_values, other_values))

    def __lt__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() > other.keys():
            raise ValueError('Slots with more keys cannot be smaller than other.')
        self_values = [self.data[k] for k in self.keys()]
        other_values = [other.data[k] for k in self.keys()]
        return (not any(s > o for s, o in zip(self_values, other_values)) and
                not (self_values == other_values))

    def __ge__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() < other.keys():
            raise ValueError('Slots with less keys cannot be larger than other.')
        self_values = [self.data[k] for k in other.keys()]
        other_values = [other.data[k] for k in other.keys()]
        return not any(s < o for s, o in zip(self_values, other_values))

    def __gt__(self, other):
        if not self.numeric or not other.numeric:
            raise TypeError('Only numeric slots can be compared.')
        if self.keys() < other.keys():
            raise ValueError('Slots with less keys cannot be larger than other.')
        self_values = [self.data[k] for k in other.keys()]
        other_values = [other.data[k] for k in other.keys()]
        return (not any(s < o for s, o in zip(self_values, other_values)) and
                not (self_values == other_values))

    # as_numeric series methods are to preserve accuracy of values.
    # as_humanized series methods are to pretty-print values.

    def as_numeric(self, slot_types, default_slot_type=None):
        data = {}
        for k, v in self.data.items():
            unit = slot_types.get(k, default_slot_type)
            if unit is None:
                raise ValueError('unit unknown for slot', k)
            data[k] = ResourceSlot.value_as_numeric(v, unit)
        return type(self)(data, numeric=True)

    @staticmethod
    def value_as_numeric(value, unit):
        if unit == 'bytes':
            if isinstance(value, (Decimal, int)):
                return value
            value = int(BinarySize.from_str(value))
        else:
            value = Decimal(value)
        return value

    @staticmethod
    def _humanize(src_data, slot_types):
        data = {}
        for k, v in src_data.items():
            unit = slot_types.get(k, 'count')
            if unit == 'bytes':
                try:
                    v = '{:s}'.format(BinarySize(v))
                except ValueError:
                    v = str(v)
            else:
                v = str(v)
            data[k] = v
        return data

    def as_humanized(self, slot_types):
        data = self._humanize(self.data, slot_types)
        return type(self)(data, numeric=False)

    def as_json_humanized(self, slot_types):
        data = self._humanize(self.data, slot_types)
        return data

    def as_json_numeric(self, slot_types, default_slot_type=None):
        data = {}
        for k, v in self.data.items():
            unit = slot_types.get(k, default_slot_type)
            if unit is None:
                raise ValueError('unit unknown for slot', k)
            data[k] = str(ResourceSlot.value_as_numeric(v, unit))
        return data

    # legacy:
    # mem: Decimal = Decimal(0)  # multiple of GiBytes
    # cpu: Decimal = Decimal(0)  # multiple of CPU cores
    # accel_slots: Optional[Mapping[str, Decimal]] = None


@attr.s(auto_attribs=True, slots=True)
class ShareRequest:
    '''
    Represents resource demands in "share" for a specific type of resource.

    **share** is a relative value where its realization is responsible to
    the specific resource type.  For example, "1.5 CUDA GPU" may be either
    physically a fraction of one CUDA GPU or fractions spanning across two (or
    more) CUDA GPUs depending on the actual GPU capacity detected/analyzed by
    the CUDA accelerator plugin.

    The version and feature set fields are optional -- they are considered
    meaningful only when specified.  If specified, they indicate the minimum
    compatible versions required to execute the given compute session.
    '''
    slot_type: SlotType
    device_share: Decimal
    feature_version: str = None
    feature_set: Set[str] = attr.Factory(set)
    lib_version: str = None
    driver_version: str = None


@attr.s(auto_attribs=True, slots=True)
class VFolderRequest:
    '''
    Represents vfolders for a new compute session request.
    '''
    vfolder_id: str
    vfolder_host: str
    permission: MountPermission


@attr.s(auto_attribs=True, slots=True)
class ResourceRequest:
    '''
    Represents resource demands for a new compute session request.

    CPU and memory shares are relative values, but you may think
    1 CPU-share is roughly equivalent to 1 CPU core and 1 memory-share is for 1
    GiB of the main memory.  Note that this mapping may be changed in the
    future depending the hardware performance changes.
    '''
    shares: Mapping[str, ShareRequest] = attr.Factory(dict)
    vfolders: Sequence[VFolderRequest] = attr.Factory(list)
    scratch_disk_size: BinarySize = None
    ignore_numa: bool = False
    extra: str = ''

    def to_json(self):
        pass

    @classmethod
    def from_json(cls):
        pass


@attr.s(auto_attribs=True, slots=True)
class SessionRequest:
    '''
    Represents a new compute session request.
    '''
    scaling_group: str
    cluster_size: int
    master_image: str
    master_resource_spec: ResourceRequest
    # for multi-container sessions
    worker_image: str = None
    worker_resource_spec: ResourceRequest = None
